@@ -850,16 +850,16 @@ class CodeEdit(QtWidgets.QPlainTextEdit):
             # selected. Except when the cursor is at the start of an otherwise
             # unselected line, because that feels unintuitive.
             has_selection = True
-            if cursor.atBlockStart():
-                cursor.movePosition(cursor.Left, cursor.MoveAnchor)
             start = cursor.selectionStart()
             end = cursor.selectionEnd()
+            if cursor.atBlockStart():
+                end -= 1
             cursor.setPosition(start, cursor.MoveAnchor)
             cursor.movePosition(cursor.StartOfBlock, cursor.MoveAnchor)
             cursor.setPosition(end, cursor.KeepAnchor)
             cursor.movePosition(cursor.EndOfBlock, cursor.KeepAnchor)
         line = cursor.selectedText().replace('\u2029', '\n')
-        if has_selection:
+        if has_selection or not line.startswith('\n'):
           line = '\n' + line
         end = cursor.selectionEnd()
         cursor.setPosition(end)
